@@ -34,12 +34,16 @@
 #include <core/dbus/types/object_path.h>
 #include <core/dbus/types/variant.h>
 
+#include <core/dbus/types/stl/tuple.h>
+
 #include <boost/utility/identity_type.hpp>
 
 #include <string>
 #include <tuple>
 #include <vector>
 
+#include <cstdint>
+
 namespace dbus = core::dbus;
 
 namespace mpris
@@ -103,6 +107,17 @@ struct Player
         static constexpr const char* stopped{"Stopped"};
     };
 
+    struct Error
+    {
+        struct OutOfProcessBufferStreamingNotSupported
+        {
+            static constexpr const char* name
+            {
+                "mpris.Player.Error.OutOfProcessBufferStreamingNotSupported"
+            };
+        };
+    };
+
     typedef std::map<std::string, core::dbus::types::Variant> Dictionary;
 
     DBUS_CPP_METHOD_DEF(Next, Player)
@@ -116,12 +131,16 @@ struct Player
     DBUS_CPP_METHOD_DEF(CreateVideoSink, Player)
     DBUS_CPP_METHOD_DEF(Key, Player)
     DBUS_CPP_METHOD_DEF(OpenUri, Player)
+    DBUS_CPP_METHOD_DEF(OpenUriExtended, Player)
 
     struct Signals
     {
         DBUS_CPP_SIGNAL_DEF(Seeked, Player, std::int64_t)
+        DBUS_CPP_SIGNAL_DEF(AboutToFinish, Player, void)
         DBUS_CPP_SIGNAL_DEF(EndOfStream, Player, void)
         DBUS_CPP_SIGNAL_DEF(PlaybackStatusChanged, Player, core::ubuntu::media::Player::PlaybackStatus)
+        DBUS_CPP_SIGNAL_DEF(VideoDimensionChanged, Player, core::ubuntu::media::video::Dimensions)
+        DBUS_CPP_SIGNAL_DEF(Error, Player, core::ubuntu::media::Player::Error)
     };
 
     struct Properties
@@ -132,6 +151,8 @@ struct Player
         DBUS_CPP_WRITABLE_PROPERTY_DEF(LoopStatus, Player, std::string)
         DBUS_CPP_WRITABLE_PROPERTY_DEF(TypedLoopStatus, Player, core::ubuntu::media::Player::LoopStatus)
         DBUS_CPP_WRITABLE_PROPERTY_DEF(AudioStreamRole, Player, core::ubuntu::media::Player::AudioStreamRole)
+        DBUS_CPP_READABLE_PROPERTY_DEF(Orientation, Player, core::ubuntu::media::Player::Orientation)
+        DBUS_CPP_WRITABLE_PROPERTY_DEF(Lifetime, Player, core::ubuntu::media::Player::Lifetime)
         DBUS_CPP_WRITABLE_PROPERTY_DEF(PlaybackRate, Player, double)
         DBUS_CPP_WRITABLE_PROPERTY_DEF(Rate, Player, double)
         DBUS_CPP_WRITABLE_PROPERTY_DEF(Shuffle, Player, bool)
@@ -191,6 +212,7 @@ struct Player
                 Properties::Duration::ValueType duration{0};
                 Properties::MinimumRate::ValueType minimum_rate{1.f};
                 Properties::MaximumRate::ValueType maximum_rate{1.f};
+                Properties::Orientation::ValueType orientation{core::ubuntu::media::Player::Orientation::rotate0};
             } defaults;
         };
 
@@ -211,6 +233,8 @@ struct Player
                   configuration.object->template get_property<Properties::LoopStatus>(),
                   configuration.object->template get_property<Properties::TypedLoopStatus>(),
                   configuration.object->template get_property<Properties::AudioStreamRole>(),
+                  configuration.object->template get_property<Properties::Orientation>(),
+                  configuration.object->template get_property<Properties::Lifetime>(),
                   configuration.object->template get_property<Properties::PlaybackRate>(),
                   configuration.object->template get_property<Properties::Shuffle>(),
                   configuration.object->template get_property<Properties::TypedMetaData>(),
@@ -223,8 +247,11 @@ struct Player
               signals
               {
                   configuration.object->template get_signal<Signals::Seeked>(),
+                  configuration.object->template get_signal<Signals::AboutToFinish>(),
                   configuration.object->template get_signal<Signals::EndOfStream>(),
                   configuration.object->template get_signal<Signals::PlaybackStatusChanged>(),
+                  configuration.object->template get_signal<Signals::VideoDimensionChanged>(),
+                  configuration.object->template get_signal<Signals::Error>(),
                   configuration.object->template get_signal<core::dbus::interfaces::Properties::Signals::PropertiesChanged>()
               }
         {
@@ -241,12 +268,20 @@ struct Player
             properties.loop_status->set(configuration.defaults.loop_status);
             properties.typed_loop_status->set(configuration.defaults.typed_loop_status);
             properties.audio_stream_role->set(core::ubuntu::media::Player::AudioStreamRole::multimedia);
+            properties.orientation->set(core::ubuntu::media::Player::Orientation::rotate0);
+            properties.lifetime->set(core::ubuntu::media::Player::Lifetime::normal);
             properties.playback_rate->set(configuration.defaults.playback_rate);
             properties.is_shuffle->set(configuration.defaults.shuffle);
             properties.position->set(configuration.defaults.position);
             properties.duration->set(configuration.defaults.duration);
             properties.minimum_playback_rate->set(configuration.defaults.minimum_rate);
-            properties.maximum_playback_rate->set(configuration.defaults.maximum_rate);            
+            properties.maximum_playback_rate->set(configuration.defaults.maximum_rate);
+
+            // Make sure the Orientation Property gets sent over DBus to the client
+            properties.orientation->changed().connect([this](const core::ubuntu::media::Player::Orientation& o)
+            {
+                on_property_value_changed<Properties::Orientation>(o);
+            });
 
             properties.position->changed().connect([this](std::int64_t position)
             {
@@ -266,7 +301,7 @@ struct Player
             properties.loop_status->changed().connect([this](const std::string& status)
             {
                 on_property_value_changed<Properties::LoopStatus>(status);
-            });                       
+            });
         }
 
         template<typename Property>
@@ -294,6 +329,8 @@ struct Player
             dict[Properties::LoopStatus::name()] = dbus::types::Variant::encode(properties.loop_status->get());
             dict[Properties::TypedLoopStatus::name()] = dbus::types::Variant::encode(properties.typed_loop_status->get());
             dict[Properties::AudioStreamRole::name()] = dbus::types::Variant::encode(properties.audio_stream_role->get());
+            dict[Properties::Orientation::name()] = dbus::types::Variant::encode(properties.orientation->get());
+            dict[Properties::Lifetime::name()] = dbus::types::Variant::encode(properties.lifetime->get());
             dict[Properties::PlaybackRate::name()] = dbus::types::Variant::encode(properties.playback_rate->get());
             dict[Properties::Shuffle::name()] = dbus::types::Variant::encode(properties.is_shuffle->get());
             dict[Properties::Duration::name()] = dbus::types::Variant::encode(properties.duration->get());
@@ -323,6 +360,8 @@ struct Player
             std::shared_ptr<core::dbus::Property<Properties::LoopStatus>> loop_status;
             std::shared_ptr<core::dbus::Property<Properties::TypedLoopStatus>> typed_loop_status;
             std::shared_ptr<core::dbus::Property<Properties::AudioStreamRole>> audio_stream_role;
+            std::shared_ptr<core::dbus::Property<Properties::Orientation>> orientation;
+            std::shared_ptr<core::dbus::Property<Properties::Lifetime>> lifetime;
             std::shared_ptr<core::dbus::Property<Properties::PlaybackRate>> playback_rate;
             std::shared_ptr<core::dbus::Property<Properties::Shuffle>> is_shuffle;
             std::shared_ptr<core::dbus::Property<Properties::TypedMetaData>> typed_meta_data_for_current_track;
@@ -336,8 +375,11 @@ struct Player
         struct
         {
             typename core::dbus::Signal<Signals::Seeked, Signals::Seeked::ArgumentType>::Ptr seeked_to;
+            typename core::dbus::Signal<Signals::AboutToFinish, Signals::AboutToFinish::ArgumentType>::Ptr about_to_finish;
             typename core::dbus::Signal<Signals::EndOfStream, Signals::EndOfStream::ArgumentType>::Ptr end_of_stream;
             typename core::dbus::Signal<Signals::PlaybackStatusChanged, Signals::PlaybackStatusChanged::ArgumentType>::Ptr playback_status_changed;
+            typename core::dbus::Signal<Signals::VideoDimensionChanged, Signals::VideoDimensionChanged::ArgumentType>::Ptr video_dimension_changed;
+            typename core::dbus::Signal<Signals::Error, Signals::Error::ArgumentType>::Ptr error;
 
             dbus::Signal
             <
