@@ -27,9 +27,6 @@
 // to the implementation in a way that is opaque to the client.
 struct core::ubuntu::media::Player::Configuration
 {
-    // An identifier that is helpful in referencing the player instance
-    // across multiple services.
-    std::string identity;
     // Unique key for identifying the session.
     core::ubuntu::media::Player::PlayerKey key;
     // The bus connection to expose objects on.
