@@ -21,11 +21,11 @@
 
 #include <unistd.h>
 
+#include "client_death_observer.h"
 #include "engine.h"
+#include "null_track_list.h"
 #include "track_list_implementation.h"
 
-#include "powerd_service.h"
-#include "unity_screen_service.h"
 #include "gstreamer/engine.h"
 
 #include <memory>
@@ -40,7 +40,8 @@ namespace dbus = core::dbus;
 
 using namespace std;
 
-struct media::PlayerImplementation::Private :
+template<typename Parent>
+struct media::PlayerImplementation<Parent>::Private :
         public std::enable_shared_from_this<Private>
 {
     enum class wakelock_clear_t
@@ -51,34 +52,40 @@ struct media::PlayerImplementation::Private :
         WAKELOCK_CLEAR_INVALID
     };
 
-    Private(PlayerImplementation* parent,
-            const dbus::types::ObjectPath& session_path,
-            const std::shared_ptr<media::Service>& service,
-            PlayerImplementation::PlayerKey key)
+    Private(PlayerImplementation* parent, const media::PlayerImplementation<Parent>::Configuration& config)
         : parent(parent),
-          service(service),
+          config(config),
+          display_state_lock(config.power_state_controller->display_state_lock()),
+          system_state_lock(config.power_state_controller->system_state_lock()),
           engine(std::make_shared<gstreamer::Engine>()),
-          session_path(session_path),
-          track_list(
-              new media::TrackListImplementation(
-                  session_path.as_string() + "/TrackList",
-                  engine->meta_data_extractor())),
-          sys_lock_name("media-hub-music-playback"),
-          disp_cookie(-1),
+          track_list(std::make_shared<NullTrackList>()),
           system_wakelock_count(0),
           display_wakelock_count(0),
           previous_state(Engine::State::stopped),
-          key(key),
-          engine_state_change_connection(engine->state().changed().connect(make_state_change_handler()))
+          engine_state_change_connection(engine->state().changed().connect(make_state_change_handler())),
+          engine_playback_status_change_connection(engine->playback_status_changed_signal().connect(make_playback_status_change_handler()))
     {
-        auto bus = std::shared_ptr<dbus::Bus>(new dbus::Bus(core::dbus::WellKnownBus::system));
-        bus->install_executor(dbus::asio::make_executor(bus));
+        std::cout << "Private parent instance: " << parent << std::endl;
+        // Poor man's logging of release/acquire events.
+        display_state_lock->acquired().connect([](media::power::DisplayState state)
+        {
+            std::cout << "Acquired new display state: " << state << std::endl;
+        });
 
-        auto stub_service = dbus::Service::use_service(bus, dbus::traits::Service<core::Powerd>::interface_name());
-        powerd_session = stub_service->object_for_path(dbus::types::ObjectPath("/com/canonical/powerd"));
+        display_state_lock->released().connect([](media::power::DisplayState state)
+        {
+            std::cout << "Released display state: " << state << std::endl;
+        });
 
-        auto uscreen_stub_service = dbus::Service::use_service(bus, dbus::traits::Service<core::UScreen>::interface_name());
-        uscreen_session = uscreen_stub_service->object_for_path(dbus::types::ObjectPath("/com/canonical/Unity/Screen"));
+        system_state_lock->acquired().connect([](media::power::SystemState state)
+        {
+            std::cout << "Acquired new system state: "  << state << std::endl;
+        });
+
+        system_state_lock->released().connect([](media::power::SystemState state)
+        {
+            std::cout << "Released system state: "  << state << std::endl;
+        });
     }
 
     ~Private()
@@ -91,6 +98,12 @@ struct media::PlayerImplementation::Private :
         // trigger the state change handler. Ensure the handler is not called
         // by disconnecting the state change signal
         engine_state_change_connection.disconnect();
+
+        std::cout << "** Disconnecting playback_status_changed_signal connection";
+        // The engine destructor can lead to a playback status change which will
+        // trigger the playback status change handler. Ensure the handler is not called
+        // by disconnecting the playback status change signal
+        engine_playback_status_change_connection.disconnect();
     }
 
     std::function<void(const Engine::State& state)> make_state_change_handler()
@@ -102,6 +115,7 @@ struct media::PlayerImplementation::Private :
          */
         return [this](const Engine::State& state)
         {
+            std::cout << "Setting state for parent: " << parent << std::endl;
             switch(state)
             {
             case Engine::State::ready:
@@ -120,6 +134,7 @@ struct media::PlayerImplementation::Private :
                 parent->meta_data_for_current_track().set(std::get<1>(engine->track_meta_data().get()));
                 // And update our playback status.
                 parent->playback_status().set(media::Player::playing);
+                std::cout << "Requesting power state" << std::endl;
                 request_power_state();
                 break;
             }
@@ -150,30 +165,36 @@ struct media::PlayerImplementation::Private :
         };
     }
 
+    std::function<void(const media::Player::PlaybackStatus& status)> make_playback_status_change_handler()
+    {
+        return [this](const media::Player::PlaybackStatus& status)
+        {
+            std::cout << "Emiting playback_status_changed for parent: " << parent << std::endl;
+            parent->emit_playback_status_changed(status);
+        };
+    }
+
     void request_power_state()
     {
+        std::cout << __PRETTY_FUNCTION__ << std::endl;
         try
         {
             if (parent->is_video_source())
             {
                 if (++display_wakelock_count == 1)
                 {
-                    auto result = uscreen_session->invoke_method_synchronously<core::UScreen::keepDisplayOn, int>();
-                    if (result.is_error())
-                        throw std::runtime_error(result.error().print());
-                    disp_cookie = result.value();
-                    cout << "Requested new display wakelock" << endl;
+                    std::cout << "Requesting new display wakelock." << std::endl;
+                    display_state_lock->request_acquire(media::power::DisplayState::on);
+                    std::cout << "Requested new display wakelock." << std::endl;
                 }
             }
             else
             {
                 if (++system_wakelock_count == 1)
                 {
-                    auto result = powerd_session->invoke_method_synchronously<core::Powerd::requestSysState, std::string>(sys_lock_name, static_cast<int>(1));
-                    if (result.is_error())
-                        throw std::runtime_error(result.error().print());
-                    sys_cookie = result.value();
-                    cout << "Requested new system wakelock" << endl;
+                    std::cout << "Requesting new system wakelock." << std::endl;
+                    system_state_lock->request_acquire(media::power::SystemState::active);
+                    std::cout << "Requested new system wakelock." << std::endl;
                 }
             }
         }
@@ -197,18 +218,16 @@ struct media::PlayerImplementation::Private :
                     // Only actually clear the system wakelock once the count reaches zero
                     if (--system_wakelock_count == 0)
                     {
-                        cout << "Clearing system wakelock" << endl;
-                        powerd_session->invoke_method_synchronously<core::Powerd::clearSysState, void>(sys_cookie);
-                        sys_cookie.clear();
+                        std::cout << "Clearing system wakelock." << std::endl;
+                        system_state_lock->request_release(media::power::SystemState::active);
                     }
                     break;
                 case wakelock_clear_t::WAKELOCK_CLEAR_DISPLAY:
                     // Only actually clear the display wakelock once the count reaches zero
                     if (--display_wakelock_count == 0)
                     {
-                        cout << "Clearing display wakelock" << endl;
-                        uscreen_session->invoke_method_synchronously<core::UScreen::removeDisplayOnRequest, void>(disp_cookie);
-                        disp_cookie = -1;
+                        std::cout << "Clearing display wakelock." << std::endl;
+                        display_state_lock->request_release(media::power::DisplayState::on);
                     }
                     break;
                 case wakelock_clear_t::WAKELOCK_CLEAR_INVALID:
@@ -250,7 +269,7 @@ struct media::PlayerImplementation::Private :
         // the execution of the functor may surpass the lifetime of this Private
         // object instance. By keeping a weak_ptr to the private object instance
         // we can check if the object is dead before calling methods on it
-        std::weak_ptr<Private> weak_self{shared_from_this()};
+        std::weak_ptr<Private> weak_self{this->shared_from_this()};
         auto wakelock_type = current_wakelock_type();
         return [weak_self, wakelock_type] {
             if (auto self = weak_self.lock())
@@ -258,61 +277,52 @@ struct media::PlayerImplementation::Private :
         };
     }
 
-    PlayerImplementation* parent;
-    std::shared_ptr<Service> service;
+    void on_client_died()
+    {
+        engine->reset();
+    }
+
+    // Our link back to our parent.
+    media::PlayerImplementation<Parent>* parent;
+    // We just store the parameters passed on construction.
+    media::PlayerImplementation<Parent>::Configuration config;
+    media::power::StateController::Lock<media::power::DisplayState>::Ptr display_state_lock;
+    media::power::StateController::Lock<media::power::SystemState>::Ptr system_state_lock;
+
     std::shared_ptr<Engine> engine;
-    dbus::types::ObjectPath session_path;
-    std::shared_ptr<TrackListImplementation> track_list;
-    std::shared_ptr<dbus::Object> powerd_session;
-    std::shared_ptr<dbus::Object> uscreen_session;
-    std::string sys_lock_name;
-    int disp_cookie;
-    std::string sys_cookie;
+    std::shared_ptr<media::NullTrackList> track_list;
     std::atomic<int> system_wakelock_count;
     std::atomic<int> display_wakelock_count;
     Engine::State previous_state;
-    PlayerImplementation::PlayerKey key;
     core::Signal<> on_client_disconnected;
     core::Connection engine_state_change_connection;
+    core::Connection engine_playback_status_change_connection;
 };
 
-media::PlayerImplementation::PlayerImplementation(
-        const std::string& identity,
-        const std::shared_ptr<core::dbus::Bus>& bus,
-        const std::shared_ptr<core::dbus::Object>& session,
-        const std::shared_ptr<Service>& service,
-        PlayerKey key)
-    : media::PlayerSkeleton
-      {
-          media::PlayerSkeleton::Configuration
-          {
-              bus,
-              session,
-              identity
-          }
-      },
-      d(make_shared<Private>(
-            this,
-            session->path(),
-            service,
-            key))
+template<typename Parent>
+media::PlayerImplementation<Parent>::PlayerImplementation(const media::PlayerImplementation<Parent>::Configuration& config)
+    : Parent{config.parent},
+      d{std::make_shared<Private>(this, config)}
 {
-    // Initializing default values for properties
-    can_play().set(true);
-    can_pause().set(true);
-    can_seek().set(true);
-    can_go_previous().set(true);
-    can_go_next().set(true);
-    is_video_source().set(false);
-    is_audio_source().set(false);
-    is_shuffle().set(true);
-    playback_rate().set(1.f);
-    playback_status().set(Player::PlaybackStatus::null);
-    loop_status().set(Player::LoopStatus::none);
-    position().set(0);
-    duration().set(0);
-    audio_stream_role().set(Player::AudioStreamRole::multimedia);
+    // Initialize default values for Player interface properties
+    Parent::can_play().set(true);
+    Parent::can_pause().set(true);
+    Parent::can_seek().set(true);
+    Parent::can_go_previous().set(true);
+    Parent::can_go_next().set(true);
+    Parent::is_video_source().set(false);
+    Parent::is_audio_source().set(false);
+    Parent::is_shuffle().set(true);
+    Parent::playback_rate().set(1.f);
+    Parent::playback_status().set(Player::PlaybackStatus::null);
+    Parent::loop_status().set(Player::LoopStatus::none);
+    Parent::position().set(0);
+    Parent::duration().set(0);
+    Parent::audio_stream_role().set(Player::AudioStreamRole::multimedia);
     d->engine->audio_stream_role().set(Player::AudioStreamRole::multimedia);
+    Parent::orientation().set(Player::Orientation::rotate0);
+    Parent::lifetime().set(Player::Lifetime::normal);
+    d->engine->lifetime().set(Player::Lifetime::normal);
 
     // Make sure that the Position property gets updated from the Engine
     // every time the client requests position
@@ -320,7 +330,7 @@ media::PlayerImplementation::PlayerImplementation(
     {
         return d->engine->position().get();
     };
-    position().install(position_getter);
+    Parent::position().install(position_getter);
 
     // Make sure that the Duration property gets updated from the Engine
     // every time the client requests duration
@@ -328,29 +338,43 @@ media::PlayerImplementation::PlayerImplementation(
     {
         return d->engine->duration().get();
     };
-    duration().install(duration_getter);
+    Parent::duration().install(duration_getter);
 
     std::function<bool()> video_type_getter = [this]()
     {
         return d->engine->is_video_source().get();
     };
-    is_video_source().install(video_type_getter);
+    Parent::is_video_source().install(video_type_getter);
 
     std::function<bool()> audio_type_getter = [this]()
     {
         return d->engine->is_audio_source().get();
     };
-    is_audio_source().install(audio_type_getter);
+    Parent::is_audio_source().install(audio_type_getter);
 
     // Make sure that the audio_stream_role property gets updated on the Engine side
     // whenever the client side sets the role
-    audio_stream_role().changed().connect([this](media::Player::AudioStreamRole new_role)
+    Parent::audio_stream_role().changed().connect([this](media::Player::AudioStreamRole new_role)
     {
         d->engine->audio_stream_role().set(new_role);
     });
 
+    // When the value of the orientation Property is changed in the Engine by playbin,
+    // update the Player's cached value
+    d->engine->orientation().changed().connect([this](const Player::Orientation& o)
+    {
+        Parent::orientation().set(o);
+    });
+
+    Parent::lifetime().changed().connect([this](media::Player::Lifetime lifetime)
+    {
+        d->engine->lifetime().set(lifetime);
+    });
+
     d->engine->about_to_finish_signal().connect([this]()
     {
+        Parent::about_to_finish()();
+
         if (d->track_list->has_next())
         {
             Track::UriType uri = d->track_list->query_uri_for_track(d->track_list->next());
@@ -370,21 +394,47 @@ media::PlayerImplementation::PlayerImplementation(
 
     d->engine->seeked_to_signal().connect([this](uint64_t value)
     {
-        seeked_to()(value);
+        Parent::seeked_to()(value);
     });
 
     d->engine->end_of_stream_signal().connect([this]()
     {
-        end_of_stream()();
+        Parent::end_of_stream()();
     });
 
-    d->engine->playback_status_changed_signal().connect([this](const Player::PlaybackStatus& status)
+    d->engine->video_dimension_changed_signal().connect([this](const media::video::Dimensions& dimensions)
     {
-        playback_status_changed()(status);
+        Parent::video_dimension_changed()(dimensions);
+    });
+
+    d->engine->error_signal().connect([this](const Player::Error& e)
+    {
+        Parent::error()(e);
+    });
+
+    // Everything is setup, we now subscribe to death notifications.
+    std::weak_ptr<Private> wp{d};
+
+    d->config.client_death_observer->register_for_death_notifications_with_key(config.key);
+    d->config.client_death_observer->on_client_with_key_died().connect([wp](const media::Player::PlayerKey& died)
+    {
+        if (auto sp = wp.lock())
+        {
+            if (died != sp->config.key)
+                return;
+
+            static const std::chrono::milliseconds timeout{1000};
+            media::timeout(timeout.count(), true, [wp]()
+            {
+                if (auto sp = wp.lock())
+                    sp->on_client_died();
+            });
+        }
     });
 }
 
-media::PlayerImplementation::~PlayerImplementation()
+template<typename Parent>
+media::PlayerImplementation<Parent>::~PlayerImplementation()
 {
     // Install null getters as these properties may be destroyed
     // after the engine has been destroyed since they are owned by the
@@ -393,96 +443,107 @@ media::PlayerImplementation::~PlayerImplementation()
     {
         return static_cast<uint64_t>(0);
     };
-    position().install(position_getter);
+    Parent::position().install(position_getter);
 
     std::function<uint64_t()> duration_getter = [this]()
     {
         return static_cast<uint64_t>(0);
     };
-    duration().install(duration_getter);
+    Parent::duration().install(duration_getter);
 
     std::function<bool()> video_type_getter = [this]()
     {
         return false;
     };
-    is_video_source().install(video_type_getter);
+    Parent::is_video_source().install(video_type_getter);
 
     std::function<bool()> audio_type_getter = [this]()
     {
         return false;
     };
-    is_audio_source().install(audio_type_getter);
+    Parent::is_audio_source().install(audio_type_getter);
 }
 
-std::shared_ptr<media::TrackList> media::PlayerImplementation::track_list()
+template<typename Parent>
+std::shared_ptr<media::TrackList> media::PlayerImplementation<Parent>::track_list()
 {
     return d->track_list;
 }
 
 // TODO: Convert this to be a property instead of sync call
-media::Player::PlayerKey media::PlayerImplementation::key() const
+template<typename Parent>
+media::Player::PlayerKey media::PlayerImplementation<Parent>::key() const
 {
-    return d->key;
+    return d->config.key;
 }
 
-bool media::PlayerImplementation::open_uri(const Track::UriType& uri)
+template<typename Parent>
+media::video::Sink::Ptr media::PlayerImplementation<Parent>::create_gl_texture_video_sink(std::uint32_t texture_id)
+{
+    d->engine->create_video_sink(texture_id);
+    return media::video::Sink::Ptr{};
+}
+
+template<typename Parent>
+bool media::PlayerImplementation<Parent>::open_uri(const Track::UriType& uri)
 {
     return d->engine->open_resource_for_uri(uri);
 }
 
-void media::PlayerImplementation::create_video_sink(uint32_t texture_id)
+template<typename Parent>
+bool media::PlayerImplementation<Parent>::open_uri(const Track::UriType& uri, const Player::HeadersType& headers)
 {
-    d->engine->create_video_sink(texture_id);
+    return d->engine->open_resource_for_uri(uri, headers);
 }
 
-media::Player::GLConsumerWrapperHybris media::PlayerImplementation::gl_consumer() const
-{
-    // This method is client-side only and is simply a no-op for the service side
-    return NULL;
-}
-
-void media::PlayerImplementation::next()
+template<typename Parent>
+void media::PlayerImplementation<Parent>::next()
 {
 }
 
-void media::PlayerImplementation::previous()
+template<typename Parent>
+void media::PlayerImplementation<Parent>::previous()
 {
 }
 
-void media::PlayerImplementation::play()
+template<typename Parent>
+void media::PlayerImplementation<Parent>::play()
 {
     d->engine->play();
 }
 
-void media::PlayerImplementation::pause()
+template<typename Parent>
+void media::PlayerImplementation<Parent>::pause()
 {
     d->engine->pause();
 }
 
-void media::PlayerImplementation::stop()
+template<typename Parent>
+void media::PlayerImplementation<Parent>::stop()
 {
     std::cout << __PRETTY_FUNCTION__ << std::endl;
     d->engine->stop();
 }
 
-void media::PlayerImplementation::set_frame_available_callback(
-    UNUSED FrameAvailableCb cb, UNUSED void *context)
-{
-    // This method is client-side only and is simply a no-op for the service side
-}
-
-void media::PlayerImplementation::set_playback_complete_callback(
-    UNUSED PlaybackCompleteCb cb, UNUSED void *context)
-{
-    // This method is client-side only and is simply a no-op for the service side
-}
-
-void media::PlayerImplementation::seek_to(const std::chrono::microseconds& ms)
+template<typename Parent>
+void media::PlayerImplementation<Parent>::seek_to(const std::chrono::microseconds& ms)
 {
     d->engine->seek_to(ms);
 }
 
-const core::Signal<>& media::PlayerImplementation::on_client_disconnected() const
+template<typename Parent>
+const core::Signal<>& media::PlayerImplementation<Parent>::on_client_disconnected() const
 {
     return d->on_client_disconnected;
 }
+
+template<typename Parent>
+void media::PlayerImplementation<Parent>::emit_playback_status_changed(const media::Player::PlaybackStatus &status)
+{
+    Parent::playback_status_changed()(status);
+}
+
+#include <core/media/player_skeleton.h>
+
+// For linking purposes, we have to make sure that we have all symbols included within the dso.
+template class media::PlayerImplementation<media::PlayerSkeleton>;
