@@ -1,5 +1,5 @@
 /*
- * Copyright © 2013-2014 Canonical Ltd.
+ * Copyright © 2013-2015 Canonical Ltd.
  *
  * This program is free software: you can redistribute it and/or modify it
  * under the terms of the GNU Lesser General Public License version 3,
@@ -17,14 +17,16 @@
  *              Jim Hodapp <jim.hodapp@canonical.com>
  */
 
-#include "apparmor.h"
 #include "codec.h"
+#include "engine.h"
+#include "external_services.h"
 #include "player_skeleton.h"
 #include "player_traits.h"
 #include "property_stub.h"
 #include "the_session_bus.h"
 #include "xesam.h"
 
+#include "apparmor/ubuntu.h"
 #include "mpris/media_player2.h"
 #include "mpris/metadata.h"
 #include "mpris/player.h"
@@ -43,24 +45,27 @@ namespace media = core::ubuntu::media;
 struct media::PlayerSkeleton::Private
 {
     Private(media::PlayerSkeleton* player,
-            const std::string& identity,
             const std::shared_ptr<core::dbus::Bus>& bus,
-            const std::shared_ptr<core::dbus::Object>& session)
+            const std::shared_ptr<core::dbus::Object>& session,
+            const apparmor::ubuntu::RequestContextResolver::Ptr& request_context_resolver,
+            const apparmor::ubuntu::RequestAuthenticator::Ptr& request_authenticator)
         : impl(player),
-          identity(identity),
           bus(bus),
           object(session),
-          apparmor_session(nullptr),
-          dbus_stub{bus},
+          request_context_resolver{request_context_resolver},
+          request_authenticator{request_authenticator},
           skeleton{mpris::Player::Skeleton::Configuration{bus, session, mpris::Player::Skeleton::Configuration::Defaults{}}},
           signals
           {
               skeleton.signals.seeked_to,
+              skeleton.signals.about_to_finish,
               skeleton.signals.end_of_stream,
-              skeleton.signals.playback_status_changed
+              skeleton.signals.playback_status_changed,
+              skeleton.signals.video_dimension_changed,
+              skeleton.signals.error
           }
     {
-    }    
+    }
 
     void handle_next(const core::dbus::Message::Ptr& msg)
     {
@@ -134,86 +139,30 @@ struct media::PlayerSkeleton::Private
     {
         uint32_t texture_id;
         in->reader() >> texture_id;
-        impl->create_video_sink(texture_id);
 
-        auto reply = dbus::Message::make_method_return(in);
+        core::dbus::Message::Ptr reply;
+
+        try
+        {
+            impl->create_gl_texture_video_sink(texture_id);
+            reply = dbus::Message::make_method_return(in);
+        }
+        catch (const media::Player::Errors::OutOfProcessBufferStreamingNotSupported& e)
+        {
+            reply = dbus::Message::make_error(
+                        in,
+                        mpris::Player::Error::OutOfProcessBufferStreamingNotSupported::name,
+                        e.what());
+        }
+        catch (...)
+        {
+            reply = dbus::Message::make_error(
+                        in,
+                        mpris::Player::Error::OutOfProcessBufferStreamingNotSupported::name,
+                        std::string{});
+        }
+
         bus->send(reply);
-    }    
-
-    bool does_client_have_access(const std::string& context, const std::string& uri)
-    {
-        if (context.empty() || uri.empty())
-        {
-            std::cout << "Client denied access since context or uri are empty" << std::endl;
-            return false;
-        }
-
-        if (context == "unconfined")
-        {
-            std::cout << "Client allowed access since it's unconfined" << std::endl;
-            return true;
-        }
-
-        size_t pos = context.find_first_of('_');
-        if (pos == std::string::npos)
-        {
-            std::cout << "Client denied access since it's an invalid apparmor security context" << std::endl;
-            return false;
-        }
-
-        const std::string pkgname = context.substr(0, pos);
-        std::cout << "client pkgname: " << pkgname << std::endl;
-        std::cout << "uri: " << uri << std::endl;
-
-        // All confined apps can access their own files
-        if (uri.find(std::string(".local/share/" + pkgname + "/")) != std::string::npos
-                || uri.find(std::string(".cache/" + pkgname + "/")) != std::string::npos)
-        {
-            std::cout << "Client can access content in ~/.local/share/" << pkgname << " or ~/.cache/" << pkgname << std::endl;
-            return true;
-        }
-        else if (uri.find(std::string("opt/click.ubuntu.com/")) != std::string::npos
-                && uri.find(pkgname) != std::string::npos)
-        {
-            std::cout << "Client can access content in own opt directory" << std::endl;
-            return true;
-        }
-        else if ((uri.find(std::string("/system/media/audio/ui/")) != std::string::npos
-                || uri.find(std::string("/android/system/media/audio/ui/")) != std::string::npos)
-                && pkgname == "com.ubuntu.camera")
-        {
-            std::cout << "Camera app can access ui sounds" << std::endl;
-            return true;
-        }
-        // TODO: Check if the trust store previously allowed direct access to uri
-
-        // Check in ~/Music and ~/Videos
-        // TODO: when the trust store lands, check it to see if this app can access the dirs and
-        // then remove the explicit whitelist of the music-app, and gallery-app
-        else if ((pkgname == "com.ubuntu.music" || pkgname == "com.ubuntu.gallery") &&
-                (uri.find(std::string("Music/")) != std::string::npos
-                || uri.find(std::string("Videos/")) != std::string::npos
-                || uri.find(std::string("/media")) != std::string::npos))
-        {
-            std::cout << "Client can access content in ~/Music or ~/Videos" << std::endl;
-            return true;
-        }
-        else if (uri.find(std::string("/usr/share/sounds")) != std::string::npos)
-        {
-            std::cout << "Client can access content in /usr/share/sounds" << std::endl;
-            return true;
-        }
-        else if (uri.find(std::string("http://")) != std::string::npos
-                || uri.find(std::string("rtsp://")) != std::string::npos)
-        {
-            std::cout << "Client can access streaming content" << std::endl;
-            return true;
-        }
-        else
-        {
-            std::cout << "Client denied access to open_uri()" << std::endl;
-            return false;
-        }
     }
 
     void handle_key(const core::dbus::Message::Ptr& in)
@@ -224,16 +173,33 @@ struct media::PlayerSkeleton::Private
     }
 
     void handle_open_uri(const core::dbus::Message::Ptr& in)
-    {        
-        dbus_stub.get_connection_app_armor_security_async(in->sender(), [this, in](const std::string& profile)
+    {
+        request_context_resolver->resolve_context_for_dbus_name_async(in->sender(), [this, in](const media::apparmor::ubuntu::Context& context)
         {
             Track::UriType uri;
             in->reader() >> uri;
 
-            bool have_access = does_client_have_access(profile, uri);
+            auto result = request_authenticator->authenticate_open_uri_request(context, uri);
 
             auto reply = dbus::Message::make_method_return(in);
-            reply->writer() << (have_access ? impl->open_uri(uri) : false);
+            reply->writer() << (std::get<0>(result) ? impl->open_uri(uri) : false);
+
+            bus->send(reply);
+        });
+    }
+
+    void handle_open_uri_extended(const core::dbus::Message::Ptr& in)
+    {
+        request_context_resolver->resolve_context_for_dbus_name_async(in->sender(), [this, in](const media::apparmor::ubuntu::Context& context)
+        {
+            Track::UriType uri;
+            Player::HeadersType headers;
+
+            in->reader() >> uri >> headers;
+
+            auto result = request_authenticator->authenticate_open_uri_request(context, uri);
+            auto reply = dbus::Message::make_method_return(in);
+            reply->writer() << (std::get<0>(result) ? impl->open_uri(uri, headers) : false);
 
             bus->send(reply);
         });
@@ -261,30 +227,39 @@ struct media::PlayerSkeleton::Private
     }
 
     media::PlayerSkeleton* impl;
-    std::string identity;
     dbus::Bus::Ptr bus;
     dbus::Object::Ptr object;
-    dbus::Object::Ptr apparmor_session;
+    media::apparmor::ubuntu::RequestContextResolver::Ptr request_context_resolver;
+    media::apparmor::ubuntu::RequestAuthenticator::Ptr request_authenticator;
 
-    org::freedesktop::dbus::DBus::Stub dbus_stub;
-
-    mpris::Player::Skeleton skeleton;    
+    mpris::Player::Skeleton skeleton;
 
     struct Signals
     {
         typedef core::dbus::Signal<mpris::Player::Signals::Seeked, mpris::Player::Signals::Seeked::ArgumentType> DBusSeekedToSignal;
         typedef core::dbus::Signal<mpris::Player::Signals::EndOfStream, mpris::Player::Signals::EndOfStream::ArgumentType> DBusEndOfStreamSignal;
+        typedef core::dbus::Signal<mpris::Player::Signals::AboutToFinish, mpris::Player::Signals::AboutToFinish::ArgumentType> DBusAboutToFinishSignal;
         typedef core::dbus::Signal<mpris::Player::Signals::PlaybackStatusChanged, mpris::Player::Signals::PlaybackStatusChanged::ArgumentType> DBusPlaybackStatusChangedSignal;
+        typedef core::dbus::Signal<mpris::Player::Signals::VideoDimensionChanged, mpris::Player::Signals::VideoDimensionChanged::ArgumentType> DBusVideoDimensionChangedSignal;
+        typedef core::dbus::Signal<mpris::Player::Signals::Error, mpris::Player::Signals::Error::ArgumentType> DBusErrorSignal;
 
         Signals(const std::shared_ptr<DBusSeekedToSignal>& remote_seeked,
+                const std::shared_ptr<DBusAboutToFinishSignal>& remote_atf,
                 const std::shared_ptr<DBusEndOfStreamSignal>& remote_eos,
-                const std::shared_ptr<DBusPlaybackStatusChangedSignal>& remote_playback_status_changed)
+                const std::shared_ptr<DBusPlaybackStatusChangedSignal>& remote_playback_status_changed,
+                const std::shared_ptr<DBusVideoDimensionChangedSignal>& remote_video_dimension_changed,
+                const std::shared_ptr<DBusErrorSignal>& remote_error)
         {
             seeked_to.connect([remote_seeked](std::uint64_t value)
             {
                 remote_seeked->emit(value);
             });
 
+            about_to_finish.connect([remote_atf]()
+            {
+                remote_atf->emit();
+            });
+
             end_of_stream.connect([remote_eos]()
             {
                 remote_eos->emit();
@@ -294,17 +269,30 @@ struct media::PlayerSkeleton::Private
             {
                 remote_playback_status_changed->emit(status);
             });
+
+            video_dimension_changed.connect([remote_video_dimension_changed](const media::video::Dimensions& dimensions)
+            {
+                remote_video_dimension_changed->emit(dimensions);
+            });
+
+            error.connect([remote_error](const media::Player::Error& e)
+            {
+                remote_error->emit(e);
+            });
         }
 
         core::Signal<int64_t> seeked_to;
+        core::Signal<void> about_to_finish;
         core::Signal<void> end_of_stream;
         core::Signal<media::Player::PlaybackStatus> playback_status_changed;
+        core::Signal<media::video::Dimensions> video_dimension_changed;
+        core::Signal<media::Player::Error> error;
     } signals;
 
 };
 
 media::PlayerSkeleton::PlayerSkeleton(const media::PlayerSkeleton::Configuration& config)
-        : d(new Private{this, config.identity, config.bus, config.session})
+        : d(new Private{this, config.bus, config.session, config.request_context_resolver, config.request_authenticator})
 {
     // Setup method handlers for mpris::Player methods.
     auto next = std::bind(&Private::handle_next, d, std::placeholders::_1);
@@ -344,6 +332,11 @@ media::PlayerSkeleton::PlayerSkeleton(const media::PlayerSkeleton::Configuration
         std::bind(&Private::handle_key,
                   d,
                   std::placeholders::_1));
+
+    d->object->install_method_handler<mpris::Player::OpenUriExtended>(
+        std::bind(&Private::handle_open_uri_extended,
+                  d,
+                  std::placeholders::_1));
 }
 
 media::PlayerSkeleton::~PlayerSkeleton()
@@ -361,6 +354,7 @@ media::PlayerSkeleton::~PlayerSkeleton()
    d->object->uninstall_method_handler<mpris::Player::OpenUri>();
    d->object->uninstall_method_handler<mpris::Player::CreateVideoSink>();
    d->object->uninstall_method_handler<mpris::Player::Key>();
+   d->object->uninstall_method_handler<mpris::Player::OpenUriExtended>();
 }
 
 const core::Property<bool>& media::PlayerSkeleton::can_play() const
@@ -443,6 +437,16 @@ const core::Property<media::Player::AudioStreamRole>& media::PlayerSkeleton::aud
     return *d->skeleton.properties.audio_stream_role;
 }
 
+const core::Property<media::Player::Orientation>& media::PlayerSkeleton::orientation() const
+{
+    return *d->skeleton.properties.orientation;
+}
+
+const core::Property<media::Player::Lifetime>& media::PlayerSkeleton::lifetime() const
+{
+    return *d->skeleton.properties.lifetime;
+}
+
 const core::Property<media::Player::PlaybackRate>& media::PlayerSkeleton::minimum_playback_rate() const
 {
     return *d->skeleton.properties.minimum_playback_rate;
@@ -488,6 +492,16 @@ core::Property<media::Player::AudioStreamRole>& media::PlayerSkeleton::audio_str
     return *d->skeleton.properties.audio_stream_role;
 }
 
+core::Property<media::Player::Orientation>& media::PlayerSkeleton::orientation()
+{
+    return *d->skeleton.properties.orientation;
+}
+
+core::Property<media::Player::Lifetime>& media::PlayerSkeleton::lifetime()
+{
+    return *d->skeleton.properties.lifetime;
+}
+
 core::Property<media::Player::PlaybackStatus>& media::PlayerSkeleton::playback_status()
 {
     return *d->skeleton.properties.typed_playback_status;
@@ -528,7 +542,6 @@ core::Property<bool>& media::PlayerSkeleton::is_audio_source()
     return *d->skeleton.properties.is_audio_source;
 }
 
-
 core::Property<media::Track::MetaData>& media::PlayerSkeleton::meta_data_for_current_track()
 {
     return *d->skeleton.properties.typed_meta_data_for_current_track;
@@ -554,6 +567,16 @@ core::Signal<int64_t>& media::PlayerSkeleton::seeked_to()
     return d->signals.seeked_to;
 }
 
+const core::Signal<void>& media::PlayerSkeleton::about_to_finish() const
+{
+    return d->signals.about_to_finish;
+}
+
+core::Signal<void>& media::PlayerSkeleton::about_to_finish()
+{
+    return d->signals.about_to_finish;
+}
+
 const core::Signal<void>& media::PlayerSkeleton::end_of_stream() const
 {
     return d->signals.end_of_stream;
@@ -568,3 +591,23 @@ core::Signal<media::Player::PlaybackStatus>& media::PlayerSkeleton::playback_sta
 {
     return d->signals.playback_status_changed;
 }
+
+const core::Signal<media::video::Dimensions>& media::PlayerSkeleton::video_dimension_changed() const
+{
+    return d->signals.video_dimension_changed;
+}
+
+core::Signal<media::video::Dimensions>& media::PlayerSkeleton::video_dimension_changed()
+{
+    return d->signals.video_dimension_changed;
+}
+
+core::Signal<media::Player::Error>& media::PlayerSkeleton::error()
+{
+    return d->signals.error;
+}
+
+const core::Signal<media::Player::Error>& media::PlayerSkeleton::error() const
+{
+    return d->signals.error;
+}
