@@ -20,6 +20,7 @@
 #define CORE_UBUNTU_MEDIA_SERVICE_IMPLEMENTATION_H_
 
 #include "service_skeleton.h"
+#include "external_services.h"
 
 namespace core
 {
@@ -27,21 +28,29 @@ namespace ubuntu
 {
 namespace media
 {
-
 class Player;
 
-class ServiceImplementation : public ServiceSkeleton
+class ServiceImplementation : public Service
 {
 public:
-    ServiceImplementation ();
+    // All creation time arguments go here.
+    struct Configuration
+    {
+        KeyedPlayerStore::Ptr player_store;
+        helper::ExternalServices& external_services;
+    };
+
+    ServiceImplementation (const Configuration& configuration);
     ~ServiceImplementation ();
 
     std::shared_ptr<Player> create_session(const Player::Configuration&);
-
+    std::shared_ptr<Player> create_fixed_session(const std::string& name, const Player::Configuration&);
+    std::shared_ptr<Player> resume_session(Player::PlayerKey key);
     void pause_other_sessions(Player::PlayerKey key);
 
 private:
-    void pause_all_multimedia_sessions();
+    void pause_all_multimedia_sessions(bool resume_play_after_phonecall);
+    void resume_paused_multimedia_sessions(bool resume_video_sessions = true);
     void resume_multimedia_session();
 
     struct Private;
